# -*- coding: utf-8 -*-

__author__ = 'Eran Kampf'
__email__ = 'eran@ekampf.com'
__version__ = '0.1.2'

from sys import getsizeof

class Maybe(object):
    pass


class Nothing(Maybe):
    def is_some(self):
        return False

    def is_none(self):
        return True

    def get(self):
        raise Exception('No such element')

    def or_else(self, els=None):
        if callable(els):
            return els()

        return els

    def __call__(self, *args, **kwargs):
        return Nothing()

    #region Comparison

    def __cmp__(self, other):
        if other.__class__ == Nothing:
            return 0

        return 1

    def __eq__(self, other):
        if other.__class__ == Nothing:
            return True

        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if other.__class__ == Nothing:
            return False

        if other.__class__ == Something:
            return True

        return True if other else False

    def __gt__(self, other):
        return False

    def __le__(self, other):
        return True

    def __ge__(self, other):
        if other.__class__ == Nothing:
            return True

        if other is None:
            return True

        return False

    #endregion

    def __getattr__(self, name):
        return Nothing()

    # region Dict
    def __len__(self):
        return 0

    def __getitem__(self, key):
        return Nothing()

    def __setitem__(self, key, value):
        pass

    def __delitem__(self, key):
        pass

    # endregion

    # region Custom representation
    def __repr__(self):
        return repr(None)

    def __str__(self):
        return str(None)

    def __unicode__(self):
        return unicode(None)

    def __nonzero__(self):
        return False

    # endregion


class Something(Maybe):
    def __init__(self, value):
        self.__value = value
        
    def __call__(self, *args, **kwargs):
        return maybe(self.__value(*args, **kwargs))

    #region Comparison
    def __cmp__(self, other):
        if other.__class__ == Nothing:
            return 1

        if other.__class__ == Something:
            return cmp(self.get(), other.get())
        else:
            return cmp(self.get(), other)

    def __eq__(self, other):
        if other.__class__ == Nothing:
            return False

        if other.__class__ == Something:
            return self.get() == other.get()

        return self.get() == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if other.__class__ == Nothing:
            return False

        if other.__class__ == Something:
            return self.get() < other.get()

        return self.get() < other


    def __gt__(self, other):
        if other.__class__ == Nothing:
            return True

        if other.__class__ == Something:
            return self.get() > other.get()

        return self.get() > other

    def __le__(self, other):
        if other.__class__ == Nothing:
            return False

        if other.__class__ == Something:
            return self.get() <= other.get()

        return self.get() <= other

    def __ge__(self, other):
        if other.__class__ == Nothing:
            return True

        if other.__class__ == Something:
            return self.get() >= other.get()

        return self.get() >= other
    #endregion

    def is_some(self):
        return True

    def is_none(self):
        return False

    def get(self):
        return self.__value

    #pylint: disable=W0613
    def or_else(self, els=None):
        return self.__value

    def __iter__(self):
        try:
            iterator = iter(self.value)
        except TypeError:
            iterator = iter([self.value])

        return iterator

    def __getattr__(self, name):
        try:
            return maybe(getattr(self.__value, name))
        except Exception:
            return Nothing()

    def __setattr__(self, name, v):
        if name == "_Something__value":
            return super(Something, self).__setattr__(name, v)

        return setattr(self.__value, name, v)

    #region Dict
    def __len__(self):
        return len(self.__value)

    def __getitem__(self, key):
        return maybe(self.__value.get(key, None))

    def __setitem__(self, key, value):
        self.__value[key] = value

    def __delitem__(self, key):
        del self.__value[key]

    #endregion

    # region Custom representation

    def __repr__(self):
        return repr(self.__value)

    def __str__(self):
        return str(self.__value)

    def __int__(self):
        return int(self.__value)

    def __long__(self):
        return long(self.__value)

    def __float__(self):
        return float(self.__value)

    def __complex__(self):
        return complex(self.__value)

    def __oct__(self):
        return oct(self.__value)

    def __hex__(self):
        return hex(self.__value)

    def __index__(self):
        return self.__value.__index__()

    def __trunc__(self):
        return self.__value.__trunc__()

    def __coerce__(self, other):
        return coerce(self.__value, other)

    def __unicode__(self):
        return unicode(self.__value)

    def __nonzero__(self):
        return True

    def __dir__(self):
        return dir(self.__value)

    def __sizeof__(self):
        return getsizeof(self.__value)
    #endregion

    #region Arithmetics
    def __add__(self, other):
        return maybe(self.__value+other)

    def __sub__(self, other):
        return maybe(self.__value-other)

    def __mul__(self, other):
        return maybe(self.__value * other)

    def __floordiv__(self, other):
        return maybe(self.__value // other)

    def __div__(self, other):
        return maybe(self.__value / other)

    def __mod__(self, other):
        return maybe(self.__value % other)

    def __divmod__(self, other):
        """Implements behavior for long division using the divmod() built in function."""
        return maybe(divmod(self.__value, other))

    def __pow__(self, other):
        """Implements behavior for exponents using the ** operator."""
        return maybe(self.__value ** other)

    def __lshift__(self, other):
        """Implements left bitwise shift using the << operator."""
        return maybe(self.__value << other)

    def __rshift__(self, other):
        """Implements right bitwise shift using the >> operator."""
        return maybe(self.__value >> other)

    def __and__(self, other):
        """Implements bitwise and using the & operator."""
        return maybe(self.__value & other)

    def __or__(self, other):
        """Implements bitwise or using the | operator."""
        return maybe(self.__value | other)

    def __xor__(self, other):
        """Implements bitwise xor using the ^ operator."""
        return maybe(self.__value ^ other)

    def __radd__(self, other):
        """Implements reflected addition."""
        return maybe(other + self.__value)

    def __rsub__(self, other):
        """Implements reflected subtraction."""
        return maybe(other - self.__value)

    def __rmul__(self, other):
        """Implements reflected multiplication."""
        return maybe(other * self.__value)

    def __rfloordiv__(self, other):
        """Implements reflected integer division using the // operator."""
        return maybe(other // self.__value)

    def __rdiv__(self, other):
        """Implements reflected division using the / operator."""
        return maybe(other / self.__value)

    def __rmod__(self, other):
        """Implements reflected modulo using the % operator."""
        return maybe(other % self.__value)

    def __rdivmod__(self, other):
        """Implements behavior for long division using the divmod() built in function, when divmod(other, self) is called."""
        return maybe(divmod(other, self.__value))

    def __rpow__(self, other):
        """Implements behavior for reflected exponents using the ** operator."""
        return maybe(other ** self.__value)

    def __rlshift__(self, other):
        """Implements reflected left bitwise shift using the << operator."""
        return maybe(other << self.__value)

    def __rrshift__(self, other):
        """Implements reflected right bitwise shift using the >> operator."""
        return maybe(other >> self.__value)

    def __rand__(self, other):
        """Implements reflected bitwise and using the & operator."""
        return maybe(other & self.__value)

    def __ror__(self, other):
        """Implements reflected bitwise or using the | operator."""
        return maybe(other | self.__value)

    def __rxor__(self, other):
        """Implements reflected bitwise xor using the ^ operator."""
        return maybe(other ^ self.__value)
    #endregion

    #region Augmented assignment
    def __iadd__(self, other):
        """Implements addition with assignment."""
        self.__value += other
        return self

    def __isub__(self, other):
        """Implements subtraction with assignment."""
        self.__value -= other
        return self

    def __imul__(self, other):
        """Implements multiplication with assignment."""
        self.__value *= other
        return self

    def __ifloordiv__(self, other):
        """Implements integer division with assignment using the //= operator."""
        self.__value //= other
        return self

    def __idiv__(self, other):
        """Implements division with assignment using the /= operator."""
        self.__value /= other
        return self

    def __imod__(self, other):
        """Implements modulo with assignment using the %= operator."""
        self.__value %= other
        return self

    def __ipow__(self, other):
        """Implements behavior for exponents with assignment using the **= operator."""
        self.__value **= other
        return self

    def __ilshift__(self, other):
        """Implements left bitwise shift with assignment using the <<= operator."""
        self.__value <<= other
        return self

    def __irshift__(self, other):
        """Implements right bitwise shift with assignment using the >>= operator."""
        self.__value >>= other
        return self

    def __iand__(self, other):
        """Implements bitwise and with assignment using the &= operator."""
        self.__value &= other
        return self

    def __ior__(self, other):
        """Implements bitwise or with assignment using the |= operator."""
        self.__value |= other
        return self

    def __ixor__(self, other):
        """Implements bitwise xor with assignment using the ^= operator."""
        self.__value ^= other
        return self

    #endregion

def maybe(value):
    """Wraps an object with a Maybe instance.

      >>> maybe("I'm a value")
      "I'm a value"

      >>> maybe(None);
      None

      Testing for value:

        >>> maybe("I'm a value").is_some()
        True
        >>> maybe("I'm a value").is_none()
        False
        >>> maybe(None).is_some()
        False
        >>> maybe(None).is_none()
        True

      Simplifying IF statements:

        >>> maybe("I'm a value").get()
        "I'm a value"

        >>> maybe("I'm a value").or_else(lambda: "No value")
        "I'm a value"

        >>> maybe(None).get()
        Traceback (most recent call last):
        ...
        Exception: No such element

        >>> maybe(None).or_else(lambda: "value")
        'value'

        >>> maybe(None).or_else("value")
        'value'

      Wrap around values from object's attributes:

        class Person(object):
            def __init__(name):
                self.eran = name

        eran = maybe(Person('eran'))

        >>> eran.name
        'eran'
        >>> eran.phone_number
        None
        >>> eran.phone_number.or_else('no phone number')
        'no phone number'

        >>> maybe(4) + 8
        12
        >>> maybe(4) - 2
        2
        >>> maybe(4) * 2
        8
        
      And methods:

        >>> maybe('VALUE').lower()
        'value'
        >>> maybe(None).invalid().method().or_else('unknwon')
        'unknwon'

      Enabled easily using NestedDictionaries without having to worry
      if a value is missing.
      For example lets assume we want to load some value from the
      following dictionary:
        nested_dict = maybe({
            'store': {
                'name': 'MyStore',
                    'departments': {
                    'sales': { 'head_count': '10' }
                }
            }
        })

        >>> nested_dict['store']['name']
        'MyStore'
        >>> nested_dict['store']['address']
        None
        >>> nested_dict['store']['address']['street'].or_else('No Address Specified')
        'No Address Specified'
        >>> nested_dict['store']['departments']['sales']['head_count'].or_else('0')
        '10'
        >>> nested_dict['store']['departments']['marketing']['head_count'].or_else('0')
        '0'

    """
    if isinstance(value, Maybe):
        return value

    if value:
        return Something(value)

    return Nothing()


def get_doctest_globs():
    class Person(object):
        def __init__(self, name):
            self.name = name

    eran = Person('eran')

    globals_dict = {
        'nested_dict': maybe({
            'store': {
                'name': 'MyStore',
                'departments': {
                    'sales': {'head_count': '10'}
                }
            }
        }),
        'eran' : maybe(eran),
        'maybe': maybe,
    }

    return globals_dict


if __name__ == "__main__":
    import doctest
    doctest.testmod(globs=get_doctest_globs())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_pymaybe
----------------------------------

Tests for `pymaybe` module.
"""

import sys
import unittest
import doctest

from pymaybe import maybe, Something, Nothing

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

def load_tests(loader, tests, ignore):
    import pymaybe
    tests.addTests(doctest.DocTestSuite(pymaybe, globs=pymaybe.get_doctest_globs()))
    return tests


class TestPyMaybe(unittest.TestCase):

    def test_maybe_withValue_returnsSomething(self):
        result = maybe("Value")
        self.assertIsInstance(result, Something)

    def test_maybe_withMaybe_returnMaybe(self):
        m = maybe("value")
        self.assertEqual(maybe(m), m)

    def test_maybe_withNone_returnsNothing(self):
        result = maybe(None)
        self.assertIsInstance(result, Nothing)


    #region Nothing - Comparison

    def test_nothing_equalToNothing(self):
        self.assertTrue(Nothing() == Nothing())

    def test_nothing_notEqualToSomething(self):
        self.assertFalse(Nothing() == Something(2))
        self.assertFalse(Something(1) == Nothing())

    def test_nothing_ltNothing_isFalse(self):
        self.assertFalse(Nothing() < Nothing())

    def test_nothing_ltSomething_isTrue(self):
        self.assertTrue(Nothing() < Something(1))

    def test_nothing_ltNone_isFalse(self):
        self.assertFalse(Nothing() < None)

    def test_nothing_ltNotNone_isFalse(self):
        self.assertTrue(Nothing() < "some")

    def test_nothing_gtAnything_isFalse(self):
        self.assertFalse(Nothing() > Nothing())
        self.assertFalse(Nothing() > Something(123))
        self.assertFalse(Nothing() > None)
        self.assertFalse(Nothing() > "Value")

    def test_nothing_leAnything_isTrue(self):
        self.assertTrue(Nothing() <= Nothing())
        self.assertTrue(Nothing() <= Something(123))
        self.assertTrue(Nothing() <= None)
        self.assertTrue(Nothing() <= "Value")

    def test_nothing_geNothing_isTrue(self):
        self.assertTrue(Nothing() >= Nothing())

    def test_nothing_geNone_isTrue(self):
        self.assertTrue(Nothing() >= None)

    def test_nothing_geNotNoneOrNothing_isFalse(self):
        self.assertFalse(Nothing() >= Something(2))
        self.assertFalse(Nothing() >= "some")

    #endregion

    #region Nothing - Misc Methods

    def test_nothing_length_isZero(self):
        self.assertEqual(len(Nothing()), 0)

    def test_nothing_getItem_returnsNone(self):
        result = Nothing()[10]
        self.assertIsInstance(result, Nothing)

    def test_nothing_strings_returnNone(self):
        n = Nothing()
        self.assertEqual(str(n), str(None))

    #endregion

    #region Something - Comparison

    def test_something_notEqualToNothing(self):
        self.assertFalse(Something(1) == Nothing())
        self.assertFalse(Nothing() == Something(2))

    def test_something_ltNothing_isFalse(self):
        self.assertFalse(Something("value") < Nothing())

    def test_something_gtNothing_isTrue(self):
        self.assertTrue(Something("value") > Nothing())

    def test_something_leNothing_isFalse(self):
        self.assertFalse(Something("value") <= Nothing())

    def test_something_geNothing_isTrue(self):
        self.assertTrue(Something("value") >= Nothing())

    #endregion

    def test_something_conversions(self):
        s = "value"
        d = dict(name="Eran")
        n = 123
        f = 3.14

        if PY2:
            self.assertEqual(unicode(Something(s)), s)

            self.assertEqual(long(Something(n)), n)
            self.assertIsInstance(long(Something(f)), long)

        self.assertEqual(str(Something(s)), s)

        self.assertEqual(repr(Something(s)), repr(s))
        self.assertEqual(repr(Something(d)),  repr(d))

        self.assertEqual(int(Something(n)), n)
        self.assertIsInstance(int(Something(n)), int)

        self.assertEqual(float(Something(f)), f)
        self.assertIsInstance(float(Something(f)), float)





    #region method call forwarding

    def test_something_forwardsMethodCalls(self):
        result = maybe('VALUE').lower()
        assert result.is_some()
        assert result == 'value', "result %s should be 'value'" % result
        assert result == maybe('value')

    def test_something_forwardsMethodCalls_handlesNonExisting(self):
        result = maybe('VALUE').lowerr()
        assert result.is_none()

    def test_nothing_forwardsMethodCalls_handlesNonExisting(self):
        result = maybe(None).invalid().call()
        assert result.is_none()

    #endregion

    # region Assertions (for compatibility between Python version)

    def assertIsInstance(self, obj, cls, msg=None):
        result = isinstance(obj, cls)
        self.assertTrue(result, msg=msg)

    #endregion


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import codecs
import os
import re

root_dir = os.path.abspath(os.path.dirname(__file__))

def get_build_number():
    fname = 'build.info'
    if os.path.isfile(fname):
        with open(fname) as f:
            build_number = f.read()
            build_number = re.sub("[^a-z0-9]+","", build_number, flags=re.IGNORECASE)
            return '.' + build_number
            
    return ''
    
def get_version(package_name):
    build_number = get_build_number()
    
    version_re = re.compile(r"^__version__ = [\"']([\w_.-]+)[\"']$")
    package_components = package_name.split('.')
    init_path = os.path.join(root_dir, *(package_components + ['__init__.py']))
    with codecs.open(init_path, 'r', 'utf-8') as f:
        for line in f:
            match = version_re.match(line[:-1])
            if match:
                return match.groups()[0]+build_number

    return '0.1.0' + build_number

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='pymaybe',
    version=get_version('pymaybe'),
    description="A Python implementation of the Maybe pattern.",
    long_description=readme + '\n\n' + history,
    author="Eran Kampf",
    author_email='eran@ekampf.com',
    url='https://github.com/ekampf/pymaybe',
    packages=[
        'pymaybe',
    ],
    package_dir={'pymaybe': 'pymaybe'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='pymaybe',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.5",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ],
    test_suite='tests',
    tests_require=test_requirements
)
